@@ -21,6 +21,7 @@
 #     https://www.nipreps.org/community/licensing/
 #
 """Filtering of :math:`B_0` field mappings with B-Splines."""
+from itertools import product
 from pathlib import Path
 import numpy as np
 import nibabel as nb
@@ -39,14 +40,13 @@ from nipype.interfaces.base import (
     OutputMultiObject,
 )
 
-from sdcflows.transform import grid_bspline_weights as gbsw
+from sdcflows.transform import grid_bspline_weights
 
 
 LOW_MEM_BLOCK_SIZE = 1000
 DEFAULT_ZOOMS_MM = (40.0, 40.0, 20.0)  # For human adults (mid-frequency), in mm
 DEFAULT_LF_ZOOMS_MM = (100.0, 100.0, 40.0)  # For human adults (low-frequency), in mm
 DEFAULT_HF_ZOOMS_MM = (16.0, 16.0, 10.0)  # For human adults (high-frequency), in mm
-BSPLINE_SUPPORT = 2 - 1.82e-3  # Disallows weights < 1e-9
 LOGGER = logging.getLogger("nipype.interface")
 
 
@@ -75,6 +75,14 @@ class _BSplineApproxInputSpec(BaseInterfaceInputSpec):
         usedefault=True,
         desc="generate a field, extrapolated outside the brain mask",
     )
+    zooms_min = traits.Union(
+        traits.Float,
+        traits.Tuple(traits.Float, traits.Float, traits.Float),
+        default_value=4.0,
+        usedefault=True,
+        desc="limit minimum image zooms, set 0.0 to use the original image",
+    )
+    debug = traits.Bool(False, usedefault=True, desc="generate extra assets for debugging")
 
 
 class _BSplineApproxOutputSpec(TraitedSpec):
@@ -124,65 +132,86 @@ class BSplineApprox(SimpleInterface):
         from sklearn import linear_model as lm
         from scipy.sparse import vstack as sparse_vstack
 
+        # Output name baseline
+        out_name = fname_presuffix(
+            self.inputs.in_data, suffix="_field", newpath=runtime.cwd
+        )
+
         # Load in the fieldmap
         fmapnii = nb.load(self.inputs.in_data)
-        data = fmapnii.get_fdata(dtype="float32")
-        mask = (
-            nb.load(self.inputs.in_mask).get_fdata() > 0
+        zooms = fmapnii.header.get_zooms()
+
+        # Get a mask (or define on the spot to cover the full extent)
+        masknii = (
+            nb.load(self.inputs.in_mask)
             if isdefined(self.inputs.in_mask)
-            else np.ones_like(data, dtype=bool)
+            else None
+        )
+
+        # Determine the shape of bspline coefficients
+        # This should not change with resizing, so do it first
+        bs_grids = [bspline_grid(fmapnii, control_zooms_mm=sp) for sp in self.inputs.bs_spacing]
+
+        need_resize = np.any(np.array(zooms) < self.inputs.zooms_min)
+        if need_resize:
+            from sdcflows.utils.tools import resample_to_zooms
+
+            target_zooms = np.maximum(zooms, self.inputs.zooms_min)
+
+            LOGGER.info(
+                "Resampling image with resolution exceeding 'zooms_min' "
+                f"({'x'.join(str(s) for s in zooms)} → "
+                f"{'x'.join(str(s) for s in target_zooms)})."
+            )
+            fmapnii = resample_to_zooms(fmapnii, target_zooms)
+
+            if masknii is not None:
+                masknii = resample_to_zooms(masknii, target_zooms)
+
+        data = fmapnii.get_fdata(dtype="float32")
+
+        # Generate a numpy array with the mask
+        mask = (
+            np.ones(fmapnii.shape, dtype=bool) if masknii is None
+            else np.asanyarray(masknii.dataobj) > 1e-4
         )
-        bs_spacing = [np.array(sp, dtype="float32") for sp in self.inputs.bs_spacing]
 
         # Recenter the fieldmap
         if self.inputs.recenter == "mode":
             from scipy.stats import mode
 
-            data -= mode(data[mask], axis=None)[0][0]
+            # Handle pre- and post-1.9 mode behavior.
+            # squeeze can be dropped when the minimum version reaches 1.9
+            # Will become: data -= mode(data[mask], keepdims=False).mode
+            data -= np.squeeze(mode(data[mask]).mode)
         elif self.inputs.recenter == "median":
             data -= np.median(data[mask])
         elif self.inputs.recenter == "mean":
             data -= np.mean(data[mask])
 
-        # Calculate the spatial location of control points
-        bs_levels = []
-        ncoeff = []
-        weights = None
-        for sp in bs_spacing:
-            level = bspline_grid(fmapnii, control_zooms_mm=sp)
-            bs_levels.append(level)
-            ncoeff.append(level.dataobj.size)
+        # Calculate collocation matrix from (possibly resized) image and knot grids
+        colmat = sparse_vstack(grid_bspline_weights(fmapnii, grid) for grid in bs_grids).T.tocsr()
 
-            weights = (
-                gbsw(fmapnii, level)
-                if weights is None
-                else sparse_vstack((weights, gbsw(fmapnii, level)))
-            )
-
-        regressors = weights.T.tocsr()[mask.reshape(-1), :]
+        bs_grids_str = ['x'.join(str(s) for s in grid.shape) for grid in bs_grids]
+        bs_grids_str[-1] = f"and {bs_grids_str[-1]}"
+        LOGGER.info(
+            f"Approximating B-Splines grids ({', '.join(bs_grids_str)} [knots]) on a grid of "
+            f"{'x'.join(str(s) for s in fmapnii.shape)} ({np.prod(fmapnii.shape)}) voxels,"
+            f" of which {mask.sum()} fall within the mask."
+        )
 
         # Fit the model
         model = lm.Ridge(alpha=self.inputs.ridge_alpha, fit_intercept=False)
-        model.fit(regressors, data[mask])
-
-        interp_data = np.zeros_like(data)
-        interp_data[mask] = np.array(model.coef_) @ regressors.T  # Interpolation
-
-        # Store outputs
-        out_name = fname_presuffix(
-            self.inputs.in_data, suffix="_field", newpath=runtime.cwd
-        )
-        hdr = fmapnii.header.copy()
-        hdr.set_data_dtype("float32")
-        fmapnii.__class__(interp_data, fmapnii.affine, hdr).to_filename(out_name)
-        self._results["out_field"] = out_name
+        model.fit(colmat[mask.reshape(-1), :], data[mask])
 
+        # Store coefficients
         index = 0
         self._results["out_coeff"] = []
-        for i, (n, bsl) in enumerate(zip(ncoeff, bs_levels)):
+        for i, bsl in enumerate(bs_grids):
+            n = bsl.dataobj.size
             out_level = out_name.replace("_field.", f"_coeff{i:03}.")
             bsl.__class__(
-                np.array(model.coef_, dtype="float32")[index : index + n].reshape(
+                np.array(model.coef_, dtype="float32")[index:index + n].reshape(
                     bsl.shape
                 ),
                 bsl.affine,
@@ -191,6 +220,29 @@ class BSplineApprox(SimpleInterface):
             index += n
             self._results["out_coeff"].append(out_level)
 
+        # Interpolating in the original grid will require a new collocation matrix
+        if need_resize:
+            fmapnii = nb.load(self.inputs.in_data)
+            data = fmapnii.get_fdata(dtype="float32")
+            mask = (
+                np.ones_like(fmapnii.dataobj, dtype=bool) if masknii is None
+                else np.asanyarray(nb.load(self.inputs.in_mask).dataobj) > 1e-4
+            )
+            colmat = sparse_vstack(
+                grid_bspline_weights(fmapnii, grid) for grid in bs_grids
+            ).T.tocsr()
+
+        regressors = colmat[mask.reshape(-1), :]
+        interp_data = np.zeros_like(data)
+        # Interpolate the field from the coefficients just calculated
+        interp_data[mask] = regressors @ model.coef_
+
+        # Store interpolated field
+        hdr = fmapnii.header.copy()
+        hdr.set_data_dtype("float32")
+        fmapnii.__class__(interp_data, fmapnii.affine, hdr).to_filename(out_name)
+        self._results["out_field"] = out_name
+
         # Write out fitting-error map
         self._results["out_error"] = out_name.replace("_field.", "_error.")
         fmapnii.__class__(
@@ -204,8 +256,8 @@ class BSplineApprox(SimpleInterface):
             self._results["out_extrapolated"] = self._results["out_field"]
             return runtime
 
-        extrapolators = weights.tocsc()[:, ~mask.reshape(-1)]
-        interp_data[~mask] = np.array(model.coef_) @ extrapolators  # Extrapolation
+        extrapolators = colmat[~mask.reshape(-1), :]
+        interp_data[~mask] = extrapolators @ model.coef_  # Extrapolation
         self._results["out_extrapolated"] = out_name.replace("_field.", "_extra.")
         fmapnii.__class__(interp_data, fmapnii.affine, hdr).to_filename(
             self._results["out_extrapolated"]
@@ -333,6 +385,7 @@ class _TransformCoefficientsInputSpec(BaseInterfaceInputSpec):
     )
     fmap_ref = File(exists=True, mandatory=True, desc="the fieldmap reference")
     transform = File(exists=True, mandatory=True, desc="rigid-body transform file")
+    fmap_target = File(exists=True, desc="the distorted EPI target (feed to set debug mode on)")
 
 
 class _TransformCoefficientsOutputSpec(TraitedSpec):
@@ -355,6 +408,10 @@ class TransformCoefficients(SimpleInterface):
                 level,
                 self.inputs.fmap_ref,
                 self.inputs.transform,
+                fmap_target=(
+                    self.inputs.fmap_target if isdefined(self.inputs.fmap_target)
+                    else None
+                ),
             )
             out_file = fname_presuffix(
                 level, suffix="_space-target", newpath=runtime.cwd
@@ -370,16 +427,9 @@ class _TOPUPCoeffReorientInputSpec(BaseInterfaceInputSpec):
     )
     fmap_ref = File(exists=True, mandatory=True, desc="the fieldmap reference")
     pe_dir = traits.Enum(
-        "+",
-        "-",
-        "i",
-        "i-",
-        "j",
-        "j-",
-        "k",
-        "k-",
-        usedefault=True,
-        desc="the polarity of the phase-encoding direction corresponding to fmap_ref",
+        *["".join(p) for p in product("ijkxyz", ("", "-"))],
+        mandatory=True,
+        desc="phase encoding direction"
     )
 
 
@@ -402,6 +452,8 @@ class TOPUPCoeffReorient(SimpleInterface):
     The q-form of these NIfTI files is always diagonal, with the decimation factors
     set on the diagonal (and hence, the voxel zooms).
     The origin of the q-form is set to the reference image's shape.
+    All the director cosines of the output coefficients will be positive.
+    In other words, the output orientation is either RAS, ARS, ASR, SAR, or SRA.
 
     This interface modifies these coefficient files to be fully-fledged NIfTI images
     aligned with the reference image.
@@ -422,7 +474,7 @@ class TOPUPCoeffReorient(SimpleInterface):
                 _fix_topup_fieldcoeff(
                     in_coeff,
                     self.inputs.fmap_ref,
-                    refpe_reversed=self.inputs.pe_dir.endswith("-"),
+                    self.inputs.pe_dir,
                     out_file=fname_presuffix(
                         in_coeff, suffix="_fixed", newpath=runtime.cwd
                     ),
@@ -461,20 +513,42 @@ def bspline_grid(img, control_zooms_mm=DEFAULT_ZOOMS_MM):
     return img.__class__(np.zeros(bs_shape, dtype="float32"), bs_affine)
 
 
-def _fix_topup_fieldcoeff(in_coeff, fmap_ref, refpe_reversed=False, out_file=None):
+def _fix_topup_fieldcoeff(in_coeff, fmap_ref, pe_dir, out_file=None):
     """Read in a coefficients file generated by TOPUP and fix x-form headers."""
     from pathlib import Path
     import numpy as np
     import nibabel as nb
+    from sdcflows.utils.tools import ensure_positive_cosines
 
     if out_file is None:
         out_file = Path("coefficients.nii.gz").absolute()
 
-    coeffnii = nb.load(in_coeff)
+    # Load reference image
     refnii = nb.load(fmap_ref)
 
+    # Load coefficients
+    coeffnii = nb.load(in_coeff)
+    # Coefficients - flip LR and overwrite coeffnii variable
+    # Internal data orientation of FSL is LAS, so coefficients will be LR flipped,
+    # and because the affine does not encode orientation (factors instead), this flip
+    # always is implicit.
+    # If the PE direction is x/i, the flip in the axis direction causes that the
+    # fieldmap estimation must also be inverted in direction (multiply by -1.0)
+    reverse_pe = -1.0 if "i" in pe_dir.replace("x", "i") else 1.0
+    lr_axis = "".join(nb.aff2axcodes(coeffnii.affine)).index("R")
+    coeffnii = coeffnii.__class__(
+        reverse_pe * np.flip(np.asanyarray(coeffnii.dataobj), axis=lr_axis),
+        coeffnii.affine,
+        coeffnii.header,
+    )
+    # Ensure reference has positive director cosines
+    refnii, ref_axcodes = ensure_positive_cosines(refnii)
+
+    # Get matrix of B-Spline control knots
     coeff_shape = np.array(coeffnii.shape[:3])
+    # Get factors (w.r.t. reference's pixel sizes) to calculate separation btw control points
     factors = np.array(coeffnii.header.get_zooms()[:3])
+    # Shape checking
     ref_shape = np.array(refnii.shape[:3])
     exp_shape = ref_shape // factors + 3 * (factors > 1)
     if not np.all(coeff_shape == exp_shape):
@@ -482,15 +556,26 @@ def _fix_topup_fieldcoeff(in_coeff, fmap_ref, refpe_reversed=False, out_file=Non
             f"Shape of coefficients file {coeff_shape} does not meet the "
             f"expected shape {exp_shape} (toupup factors are {factors})."
         )
+
+    # Contextualize the control points in space with a proper NIfTI affine
     newaff = np.eye(4)
     newaff[:3, :3] = refnii.affine[:3, :3] * factors
     c_ref = nb.affines.apply_affine(refnii.affine, 0.5 * (ref_shape - 1))
     c_coeff = nb.affines.apply_affine(newaff, 0.5 * (coeff_shape - 1))
     newaff[:3, 3] = c_ref - c_coeff
-    header = coeffnii.header.copy()
-    coeffnii.header.set_qform(coeffnii.header.get_qform(coded=False), code=0)
-    coeffnii.header.set_sform(newaff, code=1)
 
+    # Edit coefficient's header
+    header = coeffnii.header.copy()
+    header.set_qform(newaff, code=1)
+    header.set_sform(newaff, code=1)
+    header["cal_max"] = max((
+        abs(np.asanyarray(coeffnii.dataobj).min()),
+        np.asanyarray(coeffnii.dataobj).max(),
+    ))
+    header["cal_min"] = - header["cal_max"]
+    header.set_intent("estimate", tuple(), name="B-Spline coefficients")
+
+    # Write out fixed (generalized) coefficients
     coeffnii.__class__(coeffnii.dataobj, newaff, header).to_filename(out_file)
     return out_file
 
@@ -511,13 +596,13 @@ def _split_itk_file(in_file):
         yield str(p)
 
 
-def _b0_resampler(data, coeffs, pe, ro, hmc_xfm=None, unwarp=None, newpath=None):
+def _b0_resampler(in_file, coeffs, pe, ro, hmc_xfm=None, unwarp=None, newpath=None):
     """Outsource the resampler into a separate callable function to allow parallelization."""
     from functools import partial
 
     # Prepare output names
     filename = partial(fname_presuffix, newpath=newpath)
-    retval = [filename(data, suffix=s) for s in ("_unwarped", "_xfm", "_field")]
+    retval = [filename(in_file, suffix=s) for s in ("_unwarped", "_xfm", "_field")]
 
     if unwarp is None:
         from sdcflows.transform import B0FieldTransform
@@ -533,12 +618,21 @@ def _b0_resampler(data, coeffs, pe, ro, hmc_xfm=None, unwarp=None, newpath=None)
 
         unwarp.xfm = Affine(XFMLoader.from_filename(hmc_xfm).to_ras())
 
-    if unwarp.fit(data):
-        unwarp.shifts.to_filename(retval[2])
+    # Load distorted image
+    distorted_img = nb.load(in_file)
+
+    if unwarp.fit(distorted_img):
+        unwarp.mapped.to_filename(retval[2])
     else:
         retval[2] = None
 
-    unwarp.apply(nb.load(data), ro_time=ro, pe_dir=pe).to_filename(retval[0])
+    # Unwarp
+    unwarped_img = unwarp.apply(distorted_img, ro_time=ro, pe_dir=pe)
+
+    # Write out to disk
+    unwarped_img.to_filename(retval[0])
+
+    # Store the corresponding spatial transformation
     unwarp.to_displacements(ro_time=ro, pe_dir=pe).to_filename(retval[1])
 
     return retval
