@@ -23,6 +23,62 @@
 """Image processing tools."""
 
 
+def resample_to_zooms(in_file, zooms, order=3, prefilter=True):
+    """Resample the input data to a new grid with the requested zooms."""
+    from pathlib import Path
+    import numpy as np
+    import nibabel as nb
+    from nibabel.affines import rescale_affine
+    from nitransforms.linear import Affine
+
+    if isinstance(in_file, (str, Path)):
+        in_file = nb.load(in_file)
+
+    # Prepare output x-forms
+    sform, scode = in_file.get_sform(coded=True)
+    qform, qcode = in_file.get_qform(coded=True)
+
+    hdr = in_file.header.copy()
+    zooms = np.array(zooms)
+
+    pre_zooms = np.array(in_file.header.get_zooms()[:3])
+    # Could use `np.ceil` if we prefer
+    new_shape = np.rint(np.array(in_file.shape[:3]) * pre_zooms / zooms)
+    affine = rescale_affine(in_file.affine, in_file.shape[:3], zooms, new_shape)
+
+    # Generate new reference
+    hdr.set_sform(affine, scode)
+    hdr.set_qform(affine, qcode)
+    newref = in_file.__class__(
+        np.zeros(new_shape.astype(int), dtype=hdr.get_data_dtype()),
+        affine,
+        hdr,
+    )
+
+    # Resample via identity transform
+    return Affine(reference=newref).apply(in_file, order=order, prefilter=prefilter)
+
+
+def ensure_positive_cosines(img):
+    """
+    Reorient axes polarity to have all positive direction cosines.
+
+    In other words, this interface will reorient the image polarities to be all
+    *positive*, respecting the axes ordering.
+    For instance, *LAS* -> *RAS* or *ALS* -> *ARS*.
+
+    """
+    import nibabel as nb
+
+    img_axcodes = nb.aff2axcodes(img.affine)
+    in_ornt = nb.orientations.axcodes2ornt(img_axcodes)
+    out_ornt = in_ornt.copy()
+    out_ornt[:, 1] = 1
+    ornt_xfm = nb.orientations.ornt_transform(in_ornt, out_ornt)
+
+    return img.as_reoriented(ornt_xfm), img_axcodes
+
+
 def brain_masker(in_file, out_file=None, padding=5):
     """Use grayscale morphological operations to obtain a quick mask of EPI data."""
     from pathlib import Path
